@@ -1,17 +1,9 @@
 use crate::Site;
+use libs::tera::Result as TeraResult;
 use templates::{filters, global_fns};
-use tera::Result as TeraResult;
 
 /// Adds global fns that are to be available to shortcodes while rendering markdown
 pub fn register_early_global_fns(site: &mut Site) -> TeraResult<()> {
-    site.tera.register_filter(
-        "markdown",
-        filters::MarkdownFilter::new(
-            site.base_path.clone(),
-            site.config.clone(),
-            site.permalinks.clone(),
-        )?,
-    );
     site.tera.register_filter(
         "num_format",
         filters::NumFormatFilter::new(&site.config.default_language),
@@ -69,6 +61,15 @@ pub fn register_early_global_fns(site: &mut Site) -> TeraResult<()> {
         ),
     );
 
+    site.tera.register_filter(
+        "markdown",
+        filters::MarkdownFilter::new(
+            site.config.clone(),
+            site.permalinks.clone(),
+            site.tera.clone(),
+        ),
+    );
+
     Ok(())
 }
 
